#ifndef FGGL_GFX_H
#define FGGL_GFX_H

#include <cassert>
#include <string>
#include <memory>

#include <fggl/input/input.hpp>
#include <fggl/gfx/common.hpp>
#include <fggl/math/types.hpp>
#include <fggl/gfx/input.hpp>
#include <fggl/gfx/windowing.hpp>

namespace fggl::gfx {

	class GlfwContext {
		public:
			explicit GlfwContext(std::shared_ptr<fggl::input::Input> input);
			~GlfwContext();

			void pollEvents();
	};

	enum MutWindowHint {
		Decorated = GLFW_DECORATED,
		Resizable = GLFW_RESIZABLE,
		Floating = GLFW_FLOATING,
		AutoIconify = GLFW_AUTO_ICONIFY,
		FocusOnShow = GLFW_FOCUS_ON_SHOW
	};

	enum WindowHint {
		Focused = GLFW_FOCUSED,
		Iconified = GLFW_ICONIFIED,
		Maximised = GLFW_MAXIMIZED,
		Hovered = GLFW_HOVERED,
		Visible = GLFW_VISIBLE,
		Transparent = GLFW_TRANSPARENT_FRAMEBUFFER,

		GlForwardCompat = GLFW_OPENGL_FORWARD_COMPAT,
		GlDebugContext = GLFW_OPENGL_DEBUG_CONTEXT,
		NoError = GLFW_CONTEXT_NO_ERROR
	};

	class GlfwWindow : public Window {
		public:
			GlfwWindow();
			~GlfwWindow();
			GlfwWindow(Window &) = delete;
			GlfwWindow(Window &&) = delete;

			[[nodiscard]]
			math::vec2i frameSize() const override;

			[[nodiscard]]
			bool wantClose() const override;

			// window <-> opengl stuff
			void activate() const override;

			void frameStart() override;
			void frameEnd() override;

			inline float width() const {
				return m_framesize.x;
			}

			inline float height() const {
				return m_framesize.y;
			}

			inline void framesize(int width, int height) {
				m_framesize = math::vec2(width, height);
			}

			// window manager stuff
			[[nodiscard]]
			inline bool closeRequested() const {
				assert(m_window != nullptr);
				return glfwWindowShouldClose(m_window);
			}

			inline void title(const std::string &title) {
				assert(m_window != nullptr);
				glfwSetWindowTitle(m_window, title.c_str());
			}

			[[nodiscard]]
			inline bool fullscreen() const {
				assert(m_window != nullptr);
				return glfwGetWindowMonitor(m_window) != nullptr;
			}

			inline void fullscreen(bool state) {
				assert(m_window != nullptr);
				if (state) {
					auto monitor = glfwGetPrimaryMonitor();
					const auto mode = glfwGetVideoMode(monitor);
					glfwSetWindowMonitor(m_window, monitor,
										 0, 0,
										 mode->width, mode->height,
										 mode->refreshRate);
				} else {
					glfwSetWindowMonitor(m_window, nullptr,
										 0, 0,
										 800, 600, 0);
				}
			}

			[[nodiscard]]
			inline bool checkHint(WindowHint hint) const {
				return check_hint(hint);
			}

			[[nodiscard]]
			inline bool checkHint(MutWindowHint hint) const {
				return check_hint(hint);
			}

			inline void setHint(MutWindowHint hint, bool state) {
				set_hint(hint, state);
			}

			inline void visible(bool state) {
				assert(m_window != nullptr);
				if (state) {
					glfwShowWindow(m_window);
				} else {
					glfwHideWindow(m_window);
				}
			}

			inline GLFWwindow *handle() {
				assert(m_window != nullptr);
				return m_window;
			}

		private:
			GLFWwindow *m_window;
			math::vec2 m_framesize;

			inline void set_hint(int hint, bool state) const {
				assert(m_window != nullptr);
				glfwSetWindowAttrib(m_window, hint, state);
			}

			[[nodiscard]]
			inline bool check_hint(int hint) const {
				assert(m_window != nullptr);
				return glfwGetWindowAttrib(m_window, hint) == GLFW_TRUE;
			}
	};

}

#endif
